/*
    umlaut.c
    version 0.12
    umlaut-converter
    (c)1996-1998 P.Murmann

    known bugs: order of -d and -f/-D option matter!
*/

#define _POSIX_SOURCE 1
#define PROG_VERSION "umlaut 0.12 May 1998\n"

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include "umlaut.h"

static char *progname;
static char *filename = NULL;
static char
  uml_Ae = UML_Ae,
  uml_ae = UML_ae,
  uml_Oe = UML_Oe,
  uml_oe = UML_oe,
  uml_Ue = UML_Ue,
  uml_ue = UML_ue,
  uml_sz = UML_sz;
static char cmd = CMD_ASCII;
static char esc = 0;
static int cutascii = 0;
static char *outname = NULL;
static FILE *out_f = stdout;
const static char *strtab[4][7] = {
  "Ae", "ae", "Oe", "oe", "Ue", "ue", "sz",
  "&Auml;", "&auml;", "&Ouml;", "&ouml;", "&Uuml;", "&uuml;", "&szlig;",
  "\\\"A ", "\\\"a ", "\\\"O ", "\\\"o ", "\\\"U ", "\\\"u ", "\\ss ",
  "\"A", "\"a", "\"O", "\"o", "\"U", "\"u", "\"s"
  };

/*
    stringToVal():
    extract positive int-value from ASCII-string
    results:
    > -1: value
    = -1: out of range
*/

int stringToVal(char *string, int max)
{
  int i;
  unsigned long v = 0;

  while(isdigit((int)*string)) {
    v *= 10;
    v += (*string - '0'); /* ASCII-juggling */
    string++;
  }
  if(v > max)
    return(-1);
  else
    return((int)v);
}

/*
    getValFromLine():
    extract umlaut value from line
    results:
     1: OK, one value
     2: value out of range
     0: no value
    -1: syntax error
*/

int getValFromLine(char *line)
{
  int v;

  if((*line == '#') || (*line == '\n') || (*line == '\0'))
    /* possible in files */
    return(0);
  while((*line == ' ') || (*line == '\t'))
    line++;
  /* verify format 'xx:' */
  if((!isalpha((int)*line)) ||
      (!isalpha((int)*(line + 1))) ||
      (*(line + 2) != ':'))
    return(-1);
  if(!isdigit((int)*(line + 3)))
    return(-1);
  if((v = stringToVal((line + 3), 255)) == -1)
    return(2);
  *(line + 2) = '\0';
  if(strcmp(line, "Ae") == 0)
    uml_Ae = (char)v;
  else if(strcmp(line, "ae") == 0)
    uml_ae = (char)v;
  else if(strcmp(line, "Oe") == 0)
    uml_Oe = (char)v;
  else if(strcmp(line, "oe") == 0)
    uml_oe = (char)v;
  else if(strcmp(line, "Ue") == 0)
    uml_Ue = (char)v;
  else if(strcmp(line, "ue") == 0)
    uml_ue = (char)v;
  else if(strcmp(line, "sz") == 0)
    uml_sz = (char)v;
  else
    return(-1);
  return(1);
}

/*
  readDefFromFile():
  read definitions
  results:
  number of definitions read or
  -1: syntax error
*/

int readDefFromFile(char *name)
{
  int v;
  int n = 0;
  FILE *f;
  char buffer[BUFSIZ];

  if((f = fopen(name, "r")) == NULL) {
    (void)fprintf(stderr, "%s: unable to open %s\n",
      progname, name);
    exit(EXIT_FAILURE);
  }
  while(fgets((void *)buffer, (BUFSIZ - 1), f) != NULL) {
    v = getValFromLine(buffer);
    switch(v) {
    case 0:
      continue;
      break;
    case 1:
      n++;
      continue;
      break;
    case 2:
      (void)fprintf(stderr, "%s: value out of range in %s\n",
        progname, name);
      (void)fclose(f);
      return(-1);
      break;
    case -1:
      (void)fprintf(stderr, "%s: syntax error in %s\n", progname, name);
      (void)fclose(f);
      return(-1);
      break;
    default:
      break;
    }
  }
  (void)fclose(f);
  return(n);
}

/*
    openFile:
    open file and process content
    input: NULL -> stdin
    results: none
*/

void openFile(char *filename)
{
  char buffer[BUFSIZ+2];
  char *bp;
  FILE *f;
  size_t r;
  size_t total = 0;
  unsigned int line = 0;
  register int i;

  if(filename != NULL) {
    if((f = fopen(filename, "r")) == NULL) {
      (void)fprintf(stderr, "%s: unable to open %s\n",
        progname, filename);
      return;
    }
  }
  else
    f = stdin;
  while(fgets(buffer, BUFSIZ, f) != NULL) {
    line++;
    r = strlen(buffer);
    total += r;
    bp = buffer;
    for(i = 0; i < r; i++) {
      if(*bp == uml_Ae)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_Ae]);
      else if(*bp == uml_ae)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_ae]);
      else if(*bp == uml_Oe)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_Oe]);
      else if(*bp == uml_oe)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_oe]);
      else if(*bp == uml_Ue)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_Ue]);
      else if(*bp == uml_ue)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_ue]);
      else if(*bp == uml_sz)
        (void)fprintf(out_f, "%s", strtab[cmd][ORD_sz]);
      else {
	if(esc) {
	  switch(cmd) {
	  case CMD_HTML:
	    switch(*bp) {
	    case '&':
	      (void)fprintf(out_f, "&amp;");
	      bp++;
	      continue;
	      break;
	    case '<':
	      (void)fprintf(out_f, "&lt;");
	      bp++;
	      continue;
	      break;
	    case '>':
	      (void)fprintf(out_f, "&gt;");
	      bp++;
	      continue;
	      break;
	    default:
	      break;
	    }
	    break;
	  case CMD_TEX:
	  case CMD_LATEX:
	    switch(*bp) {
	    case '"':
	      (void)fprintf(out_f, "''");
	      bp++;
	      continue;
	      break;
	    case '$':
	      (void)fprintf(out_f, "\\$");
	      bp++;
	      continue;
	      break;
	    case '%':
	      (void)fprintf(out_f, "\\%");
	      bp++;
	      continue;
	      break;
	    case '&':
	      (void)fprintf(out_f, "\\&");
	      bp++;
	      continue;
	      break;
	    case '\\':
	      (void)fprintf(out_f, "\\\\");
	      bp++;
	      continue;
	      break;
	    case '^':
	      (void)fprintf(out_f, "\\^");
	      bp++;
	      continue;
	      break;
	    case '_':
	      (void)fprintf(out_f, "\\_");
	      bp++;
	      continue;
	      break;
	    case '|':
	      (void)fprintf(out_f, "$\\vert$");
	      bp++;
	      continue;
	      break;
	    case '{':
	      (void)fprintf(out_f, "\\{");
	      bp++;
	      continue;
	      break;
	    case '}':
	      (void)fprintf(out_f, "\\}");
	      bp++;
	      continue;
	      break;
	    default:
	      break;
	    }
	    break;
	  default:
	    break;
	  }
	}
        if(cutascii && ((unsigned char)*bp > 127)) { 
          bp++;   
          continue;
        }
        (void)fputc((int)*bp, out_f); 
      }
      bp++;   
    }
  }
  if(ferror(f))
    (void)fprintf(stderr, "%s: %s: error reading line %d, byte %d\n",
      progname, filename, (line + 1), (total + 1));
  (void)fclose(f);
}

/*
    getFileName():
    resolve $HOME-dir
    results:
    != NULL: path
    == NULL: error
*/

char *getFileName()
{
  char *str;
  char *name;

  if((str = getenv("HOME")) == NULL)
    return(NULL);
  if((name = (char *)malloc(strlen(str) + 8 + 1))
      == NULL)
    return(NULL);
  memcpy(name, str, (strlen(str) + 1));
  strcat(name, "/.umlaut");
  return(name);
}

void usage()
{
  (void)fprintf(stderr, "usage: %s [<option{s}>] [<file>]\n\
options are:
-a       umlaut to ASCII (default)\n\
-c       cut off non-ASCII characters\n\
-d u:v   define umlaut <u> as value <v>\n\
-D       use ~/.umlaut for definitions\n\
-e       replace escape sequences\n\
-f file  use <file> for definitions\n\
-h       this output\n\
-l       umlaut to LaTeX\n\
-m       umlaut to HTML\n\
-o file  write output to <file>\n\
-t       umlaut to TeX\n\
-V       version information\n", progname);
  exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
  int i,k;
  int opt;
  int def = 0;
  char *defname = NULL;

  progname = argv[0];
  while((opt = getopt(argc, (char * const)argv, "acd:eDf:hlmo:tV"))
  != EOF) {
    switch(opt) {
    case 'V':
      (void)printf(PROG_VERSION);
      exit(EXIT_SUCCESS);
    case 'h':
      usage();
      break;
    case 'a':
      cmd = CMD_ASCII;
      break;
    case 'm':
      cmd = CMD_HTML;
      break;
    case 'l':
      cmd = CMD_LATEX;
      break;
    case 't':
      cmd = CMD_TEX;
      break;
    case 'c':
      cutascii = 1;
      break;
    case 'd':
      if(getValFromLine(optarg) != 1)
        usage();
      def = 1;
      break;
    case 'D':
      if(def)
        (void)fprintf(stderr, "%s: warning: -D overwrites -d\n",
	  progname);
      if((defname = getFileName()) == NULL) {
	(void)fprintf(stderr, "%s: unable to find $HOME\n",
	  progname);
        exit(EXIT_FAILURE);
      }
      break;
    case 'e':
      esc = 1;
      break;
    case 'f':
      if(def)
        (void)fprintf(stderr, "%s: warning: -f overwrites -d\n",
	  progname);
      defname = optarg;
      break;
    case 'o':
      outname = optarg;
      break;
    default:
      usage();
      break;
    }
  }
  if(defname != NULL) {
    if(readDefFromFile(defname) < 0)
      exit(EXIT_FAILURE);
  }
  if(outname != NULL) {
    if((out_f = fopen(outname, "w")) == NULL) {
      (void)fprintf(stderr, "%s: unable to open %s\n",
        progname, outname);
      exit(EXIT_FAILURE);
    }
  }
  if(argv[optind] == NULL)
    openFile(NULL);
  else {
    while(argv[optind] != NULL)
      openFile(argv[optind++]);
  }
  if(out_f != NULL)
    (void)fclose(out_f);
  exit(EXIT_SUCCESS);
}
